// copyright ############################### //
// This file is part of the Xtrack Package.  //
// Copyright (c) CERN, 2023.                 //
// ######################################### //

#ifndef XTRACK_PARTICLE_STATES_H
#define XTRACK_PARTICLE_STATES_H

#define  XT_LOST_ON_APERTURE       0
#define  XT_LOST_ON_LONG_CUT      -2
#define  XT_LOST_ALL_E_IN_SYNRAD -10
#define  RNG_ERR_SEEDS_NOT_SET   -20
#define  RNG_ERR_INVALID_TRACK   -21
#define  RNG_ERR_RUTH_NOT_SET    -22

#endif /* XTRACK_PARTICLE_STATES_H */
